@@ -25,8 +25,10 @@ class BaseTestCase(testtools.TestCase):
 
 
 class FakeRequestResponse(object):
-    def __init__(self, status_code=None, json=None, exception_msg=None):
+    def __init__(self, status_code=None, json=None, text=None,
+                 exception_msg=None):
         self._json = json
+        self.text = text
         self.status_code = status_code
         self.exception_msg = exception_msg or 'Error'
 
