@@ -17,7 +17,7 @@ from tests.unit import FakeRequestResponse
 
 from unittest.mock import MagicMock
 
-from zuulclient.api import ZuulRESTClient
+from zuulclient.api import ZuulRESTClient, BearerAuth
 from zuulclient.api import ZuulRESTException
 
 
@@ -42,8 +42,8 @@ class TestApi(BaseTestCase):
         self.assertEqual('https://fake.zuul/', client.url)
         self.assertEqual('https://fake.zuul/api/', client.base_url)
         self.assertEqual(True, client.session.verify)
-        self.assertEqual('Bearer %s' % token,
-                         client.session.headers.get('Authorization'))
+        self.assertTrue(isinstance(client.session.auth, BearerAuth))
+        self.assertEqual(token, client.session.auth._token)
 
     def _test_status_check(self, client, verb, func, *args, **kwargs):
         # validate request errors
@@ -73,6 +73,7 @@ class TestApi(BaseTestCase):
 
         client = ZuulRESTClient(url='https://fake.zuul/',
                                 auth_token='aiaiaiai')
+        client.info_ = {}
         # test status checks
         self._test_status_check(
             client, 'post', client.autohold,
@@ -92,10 +93,24 @@ class TestApi(BaseTestCase):
                   'node_hold_expiration': 3600}
         )
         self.assertEqual(True, ah)
+        client.info_ = {'tenant': 'tenant1'}
+        ah = client.autohold(
+            'tenant1', 'project', 'job', 1, None, 'reason', 1, 3600)
+        client.session.post.assert_called_with(
+            'https://fake.zuul/api/project/project/autohold',
+            json={'reason': 'reason',
+                  'count': 1,
+                  'job': 'job',
+                  'change': 1,
+                  'ref': None,
+                  'node_hold_expiration': 3600}
+        )
+        self.assertEqual(True, ah)
 
     def test_autohold_list(self):
         """Test autohold-list"""
         client = ZuulRESTClient(url='https://fake.zuul/')
+        client.info_ = {}
         # test status checks
         self._test_status_check(
             client, 'get', client.autohold_list, 'tenant1')
@@ -117,16 +132,23 @@ class TestApi(BaseTestCase):
         client.session.get.assert_called_with(
             'https://fake.zuul/api/tenant/tenant1/autohold')
         self.assertEqual(fakejson, ahl)
+        client.info_ = {'tenant': 'tenant1'}
+        ahl = client.autohold_list('tenant1')
+        client.session.get.assert_called_with(
+            'https://fake.zuul/api/autohold')
+        self.assertEqual(fakejson, ahl)
 
     def test_autohold_delete(self):
         """Test autohold-delete"""
         client = ZuulRESTClient(url='https://fake.zuul/')
+        client.info_ = {}
         # token required
         with self.assertRaisesRegex(Exception, 'Auth Token required'):
             client.autohold_delete(123, 'tenant1')
 
         client = ZuulRESTClient(url='https://fake.zuul/',
                                 auth_token='aiaiaiai')
+        client.info_ = {}
         # test status checks
         self._test_status_check(
             client, 'delete', client.autohold_delete,
@@ -140,10 +162,17 @@ class TestApi(BaseTestCase):
             'https://fake.zuul/api/tenant/tenant1/autohold/123'
         )
         self.assertEqual(True, ahd)
+        client.info_ = {'tenant': 'tenant1'}
+        ahd = client.autohold_delete(123, 'tenant1')
+        client.session.delete.assert_called_with(
+            'https://fake.zuul/api/autohold/123'
+        )
+        self.assertEqual(True, ahd)
 
     def test_autohold_info(self):
         """Test autohold-info"""
         client = ZuulRESTClient(url='https://fake.zuul/')
+        client.info_ = {}
         # test status checks
         self._test_status_check(
             client, 'get', client.autohold_info, 123, 'tenant1')
@@ -165,16 +194,23 @@ class TestApi(BaseTestCase):
         client.session.get.assert_called_with(
             'https://fake.zuul/api/tenant/tenant1/autohold/123')
         self.assertEqual(fakejson, ahl)
+        client.info_ = {'tenant': 'tenant1'}
+        ahl = client.autohold_info(tenant='tenant1', id=123)
+        client.session.get.assert_called_with(
+            'https://fake.zuul/api/autohold/123')
+        self.assertEqual(fakejson, ahl)
 
     def test_enqueue(self):
         """Test enqueue"""
         client = ZuulRESTClient(url='https://fake.zuul/')
+        client.info_ = {}
         # token required
         with self.assertRaisesRegex(Exception, 'Auth Token required'):
             client.enqueue('tenant1', 'check', 'project1', '1,1')
 
         client = ZuulRESTClient(url='https://fake.zuul/',
                                 auth_token='aiaiaiai')
+        client.info_ = {}
         # test status checks
         self._test_status_check(
             client, 'post', client.enqueue,
@@ -190,10 +226,19 @@ class TestApi(BaseTestCase):
                   'pipeline': 'check'}
         )
         self.assertEqual(True, enq)
+        client.info_ = {'tenant': 'tenant1'}
+        enq = client.enqueue('tenant1', 'check', 'project1', '1,1')
+        client.session.post.assert_called_with(
+            'https://fake.zuul/api/project/project1/enqueue',
+            json={'change': '1,1',
+                  'pipeline': 'check'}
+        )
+        self.assertEqual(True, enq)
 
     def test_enqueue_ref(self):
         """Test enqueue ref"""
         client = ZuulRESTClient(url='https://fake.zuul/')
+        client.info_ = {}
         # token required
         with self.assertRaisesRegex(Exception, 'Auth Token required'):
             client.enqueue_ref(
@@ -201,6 +246,7 @@ class TestApi(BaseTestCase):
 
         client = ZuulRESTClient(url='https://fake.zuul/',
                                 auth_token='aiaiaiai')
+        client.info_ = {}
         # test status checks
         self._test_status_check(
             client, 'post', client.enqueue_ref,
@@ -219,16 +265,29 @@ class TestApi(BaseTestCase):
                   'pipeline': 'check'}
         )
         self.assertEqual(True, enq_ref)
+        client.info_ = {'tenant': 'tenant1'}
+        enq_ref = client.enqueue_ref(
+            'tenant1', 'check', 'project1', 'refs/heads/stable', '0', '0')
+        client.session.post.assert_called_with(
+            'https://fake.zuul/api/project/project1/enqueue',
+            json={'ref': 'refs/heads/stable',
+                  'oldrev': '0',
+                  'newrev': '0',
+                  'pipeline': 'check'}
+        )
+        self.assertEqual(True, enq_ref)
 
     def test_dequeue(self):
         """Test dequeue"""
         client = ZuulRESTClient(url='https://fake.zuul/')
+        client.info_ = {}
         # token required
         with self.assertRaisesRegex(Exception, 'Auth Token required'):
             client.dequeue('tenant1', 'check', 'project1', '1,1')
 
         client = ZuulRESTClient(url='https://fake.zuul/',
                                 auth_token='aiaiaiai')
+        client.info_ = {}
         # test status checks
         self._test_status_check(
             client, 'post', client.dequeue,
@@ -257,16 +316,34 @@ class TestApi(BaseTestCase):
                   'pipeline': 'check'}
         )
         self.assertEqual(True, deq)
+        client.info_ = {'tenant': 'tenant1'}
+        deq = client.dequeue('tenant1', 'check', 'project1', change='1,1')
+        client.session.post.assert_called_with(
+            'https://fake.zuul/api/project/project1/dequeue',
+            json={'change': '1,1',
+                  'pipeline': 'check'}
+        )
+        self.assertEqual(True, deq)
+        deq = client.dequeue(
+            'tenant1', 'check', 'project1', ref='refs/heads/stable')
+        client.session.post.assert_called_with(
+            'https://fake.zuul/api/project/project1/dequeue',
+            json={'ref': 'refs/heads/stable',
+                  'pipeline': 'check'}
+        )
+        self.assertEqual(True, deq)
 
     def test_promote(self):
         """Test promote"""
         client = ZuulRESTClient(url='https://fake.zuul/')
+        client.info_ = {}
         # token required
         with self.assertRaisesRegex(Exception, 'Auth Token required'):
             client.promote('tenant1', 'check', ['1,1', '2,1'])
 
         client = ZuulRESTClient(url='https://fake.zuul/',
                                 auth_token='aiaiaiai')
+        client.info_ = {}
         # test status checks
         self._test_status_check(
             client, 'post', client.promote,
@@ -282,6 +359,14 @@ class TestApi(BaseTestCase):
                   'pipeline': 'check'}
         )
         self.assertEqual(True, prom)
+        client.info_ = {'tenant': 'tenant1'}
+        prom = client.promote('tenant1', 'check', ['1,1', '2,1'])
+        client.session.post.assert_called_with(
+            'https://fake.zuul/api/promote',
+            json={'changes': ['1,1', '2,1'],
+                  'pipeline': 'check'}
+        )
+        self.assertEqual(True, prom)
 
     def test_get_key(self):
         """Test getting a project's public key"""
@@ -302,6 +387,72 @@ GuS6/ewjS+arA1Iyeg/IxmECAwEAAQ==
 -----END PUBLIC KEY-----"""
         req = FakeRequestResponse(200, text=pubkey)
         client = ZuulRESTClient(url='https://fake.zuul/')
+        client.info_ = {}
         client.session.get = MagicMock(return_value=req)
         key = client.get_key('tenant1', 'project1')
+        client.session.get.assert_called_with(
+            'https://fake.zuul/api/tenant/tenant1/key/project1.pub'
+        )
         self.assertEqual(pubkey, key)
+        client.info_ = {'tenant': 'tenant1'}
+        key = client.get_key('tenant1', 'project1')
+        client.session.get.assert_called_with(
+            'https://fake.zuul/api/key/project1.pub'
+        )
+        self.assertEqual(pubkey, key)
+
+    def test_build(self):
+        """Test build endpoint"""
+        client = ZuulRESTClient(url='https://fake.zuul/')
+        # test status checks
+        self._test_status_check(
+            client, 'get', client.build, 'tenant1', 'a1a1a1a1')
+
+        fakejson = {
+            'uuid': 'a1a1a1a1',
+            'job_name': 'tox-py38',
+            'result': 'SUCCESS',
+            'held': False,
+            'start_time': '2020-09-10T14:08:55',
+            'end_time': '2020-09-10T14:13:35',
+            'duration': 280.0,
+            'voting': True,
+            'log_url': 'https://log.storage/',
+            'node_name': None,
+            'error_detail': None,
+            'final': True,
+            'artifacts': [
+                {'name': 'Download all logs',
+                 'url': 'https://log.storage/download-logs.sh',
+                 'metadata': {
+                     'command': 'xxx'}
+                 },
+                {'name': 'Zuul Manifest',
+                 'url': 'https://log.storage/zuul-manifest.json',
+                 'metadata': {
+                     'type': 'zuul_manifest'
+                 }
+                 },
+                {'name': 'Unit Test Report',
+                 'url': 'https://log.storage/testr_results.html',
+                 'metadata': {
+                     'type': 'unit_test_report'
+                 }
+                 }],
+            'provides': [],
+            'project': 'project1',
+            'branch': 'master',
+            'pipeline': 'check',
+            'change': 1234,
+            'patchset': '1',
+            'ref': 'refs/changes/34/1234/1',
+            'newrev': None,
+            'ref_url': 'https://gerrit/1234',
+            'event_id': '6b28762adfce415ba47e440c365ae624',
+            'buildset': {'uuid': 'b1b1b1'}}
+        req = FakeRequestResponse(200, fakejson)
+        client.session.get = MagicMock(return_value=req)
+        ahl = client.build(tenant='tenant1', uuid='a1a1a1a1')
+        client.session.get.assert_any_call(
+            'https://fake.zuul/api/tenant/tenant1/build/a1a1a1a1')
+        self.assertEqual(fakejson, ahl)
