@@ -36,6 +36,17 @@ chunks = [
     '1tlAuq+vIhlY3iYlzVtPTiIOtF/6V+qPHnq1k6Tiv8YzJms1WyOuw106Bzl9XM=']
 
 
+def mock_get(func=MagicMock(return_value=None), info={}):
+
+    def funk(*args, **kwargs):
+        if args[0].endswith('/info'):
+            return FakeRequestResponse(200, info)
+        else:
+            return func(*args, **kwargs)
+
+    return funk
+
+
 class TestCmd(BaseTestCase):
 
     def test_client_args_errors(self):
@@ -52,6 +63,112 @@ class TestCmd(BaseTestCase):
                       '--reason', 'some reason',
                       '--node-hold-expiration', '3600'])
 
+    def test_use_conf(self):
+        """Test that CLI can use a configuration file"""
+        ZC = ZuulClient()
+        with tempfile.NamedTemporaryFile(delete=False) as conf_file:
+            conf_file.write(
+                b"""
+[confA]
+url=https://my.fake.zuul/
+tenant=mytenant
+auth_token=mytoken
+verify_ssl=True"""
+            )
+            conf_file.close()
+            with patch("requests.Session") as mock_sesh:
+                session = mock_sesh.return_value
+                session.post = MagicMock(
+                    return_value=FakeRequestResponse(200, True)
+                )
+                session.get = MagicMock(side_effect=mock_get())
+                exit_code = ZC._main(
+                    [
+                        "-c",
+                        conf_file.name,
+                        "--use-config",
+                        "confA",
+                        "autohold",
+                        "--project",
+                        "project1",
+                        "--job",
+                        "job1",
+                        "--change",
+                        "3",
+                        "--reason",
+                        "some reason",
+                        "--node-hold-expiration",
+                        "3600",
+                    ]
+                )
+                self.assertEqual("mytoken", ZC.get_client().auth_token)
+                self.assertEqual(True, ZC.get_client().verify)
+                session.post.assert_called_with(
+                    "https://my.fake.zuul/api/tenant/mytenant/"
+                    "project/project1/autohold",
+                    json={
+                        "reason": "some reason",
+                        "count": 1,
+                        "job": "job1",
+                        "change": "3",
+                        "ref": "",
+                        "node_hold_expiration": 3600,
+                    },
+                )
+                self.assertEqual(0, exit_code)
+        os.unlink(conf_file.name)
+
+    def test_tenant_scoping_errors(self):
+        """Test the right uses of --tenant"""
+        ZC = ZuulClient()
+        with patch('requests.Session') as mock_sesh:
+            session = mock_sesh.return_value
+            test_args = [
+                ['autohold',
+                 '--project', 'project1',
+                 '--job', 'job1', '--change', '3',
+                 '--reason', 'some reason',
+                 '--node-hold-expiration', '3600'],
+                ['autohold-delete', '1234'],
+                ['autohold-info', '1234'],
+                ['enqueue',
+                 '--pipeline', 'check',
+                 '--change', '3,1',
+                 '--project', 'project1'],
+                ['enqueue-ref',
+                 '--pipeline', 'check',
+                 '--ref', 'refs/heads/stable',
+                 '--project', 'project1',
+                 '--oldrev', 'ababababab'],
+                ['dequeue',
+                 '--pipeline', 'check',
+                 '--change', '3,3',
+                 '--project', 'project1'],
+                ['promote',
+                 '--pipeline', 'gate',
+                 '--changes', '3,3', '4,1', '5,3'],
+                ['encrypt', '--project', 'project1']
+            ]
+            for args in test_args:
+                session.get = MagicMock(
+                    side_effect=mock_get()
+                )
+                with self.assertRaisesRegex(
+                    Exception,
+                    "the --tenant argument or the 'tenant' field "
+                    "in the configuration file is required",
+                ):
+                    ZC._main(['--zuul-url', 'https://fake.zuul',
+                              '--auth-token', 'aiaiaiai', ] + args)
+                session.get = MagicMock(
+                    side_effect=mock_get(info={'info': {'tenant': 'scoped'}})
+                )
+                with self.assertRaisesRegex(Exception,
+                                            'scoped to tenant "scoped"'):
+                    ZC._main(['--zuul-url', 'https://fake.zuul',
+                              '--auth-token', 'aiaiaiai', ] + args +
+                             ['--tenant', 'tenant-' + args[0]])
+
     def test_autohold(self):
         """Test autohold via CLI"""
         ZC = ZuulClient()
@@ -59,6 +176,9 @@ class TestCmd(BaseTestCase):
             session = mock_sesh.return_value
             session.post = MagicMock(
                 return_value=FakeRequestResponse(200, True))
+            session.get = MagicMock(
+                side_effect=mock_get()
+            )
             exit_code = ZC._main(
                 ['--zuul-url', 'https://fake.zuul',
                  '--auth-token', 'aiaiaiai', 'autohold',
@@ -76,6 +196,27 @@ class TestCmd(BaseTestCase):
                       'node_hold_expiration': 3600}
             )
             self.assertEqual(0, exit_code)
+            # test scoped
+            session.get = MagicMock(
+                side_effect=mock_get(info={'info': {'tenant': 'scoped'}})
+            )
+            exit_code = ZC._main(
+                ['--zuul-url', 'https://scoped.zuul',
+                 '--auth-token', 'aiaiaiai', 'autohold',
+                 '--project', 'project1',
+                 '--job', 'job1', '--change', '3', '--reason', 'some reason',
+                 '--node-hold-expiration', '3600'])
+            session.post.assert_called_with(
+                'https://scoped.zuul/api/'
+                'project/project1/autohold',
+                json={'reason': 'some reason',
+                      'count': 1,
+                      'job': 'job1',
+                      'change': '3',
+                      'ref': '',
+                      'node_hold_expiration': 3600}
+            )
+            self.assertEqual(0, exit_code)
 
     def test_autohold_args_errors(self):
         """Test wrong arguments for autohold"""
@@ -130,6 +271,9 @@ class TestCmd(BaseTestCase):
             session = mock_sesh.return_value
             session.delete = MagicMock(
                 return_value=FakeRequestResponse(204))
+            session.get = MagicMock(
+                side_effect=mock_get()
+            )
             exit_code = ZC._main(
                 ['--zuul-url', 'https://fake.zuul',
                  '--auth-token', 'aiaiaiai', 'autohold-delete',
@@ -143,8 +287,9 @@ class TestCmd(BaseTestCase):
         ZC = ZuulClient()
         with patch('requests.Session') as mock_sesh:
             session = mock_sesh.return_value
-            session.get = MagicMock(
-                return_value=FakeRequestResponse(
+
+            def rv(*args, **kargs):
+                return FakeRequestResponse(
                     200,
                     json={'id': 1234,
                           'tenant': 'tenant1',
@@ -156,7 +301,15 @@ class TestCmd(BaseTestCase):
                           'node_expiration': 0,
                           'expired': 0,
                           'reason': 'some_reason',
-                          'nodes': ['node1', 'node2']}))
+                          'nodes': [{'build': 'alalala',
+                                     'nodes': ['node1',
+                                               'node2']}
+                                    ],
+                          })
+
+            session.get = MagicMock(
+                side_effect=mock_get(rv)
+            )
             exit_code = ZC._main(
                 ['--zuul-url', 'https://fake.zuul', 'autohold-info',
                  '--tenant', 'tenant1', '1234'])
@@ -178,6 +331,9 @@ class TestCmd(BaseTestCase):
             session = mock_sesh.return_value
             session.post = MagicMock(
                 return_value=FakeRequestResponse(200, True))
+            session.get = MagicMock(
+                side_effect=mock_get()
+            )
             exit_code = ZC._main(
                 ['--zuul-url', 'https://fake.zuul',
                  '--auth-token', 'aiaiaiai', 'enqueue',
@@ -199,6 +355,9 @@ class TestCmd(BaseTestCase):
             session = mock_sesh.return_value
             session.post = MagicMock(
                 return_value=FakeRequestResponse(200, True))
+            session.get = MagicMock(
+                side_effect=mock_get()
+            )
             # ensure default revs are set
             exit_code = ZC._main(
                 ['--zuul-url', 'https://fake.zuul',
@@ -272,6 +431,9 @@ class TestCmd(BaseTestCase):
             session = mock_sesh.return_value
             session.post = MagicMock(
                 return_value=FakeRequestResponse(200, True))
+            session.get = MagicMock(
+                side_effect=mock_get()
+            )
             exit_code = ZC._main(
                 ['--zuul-url', 'https://fake.zuul',
                  '--auth-token', 'aiaiaiai', 'dequeue',
@@ -306,6 +468,9 @@ class TestCmd(BaseTestCase):
             session = mock_sesh.return_value
             session.post = MagicMock(
                 return_value=FakeRequestResponse(200, True))
+            session.get = MagicMock(
+                side_effect=mock_get()
+            )
             exit_code = ZC._main(
                 ['--zuul-url', 'https://fake.zuul',
                  '--auth-token', 'aiaiaiai', 'promote',
@@ -329,8 +494,13 @@ class TestCmd(BaseTestCase):
         ZC = ZuulClient()
         with patch('requests.Session') as mock_sesh:
             session = mock_sesh.return_value
+
+            def rv(*args, **kwargs):
+                return FakeRequestResponse(200, text='aaa')
+
             session.get = MagicMock(
-                return_value=FakeRequestResponse(200, text='aaa'))
+                side_effect=mock_get(rv)
+            )
             with patch('zuulclient.cmd.encrypt_with_openssl') as m_encrypt:
                 m_encrypt.return_value = chunks
                 exit_code = ZC._main(
@@ -360,3 +530,100 @@ class TestCmd(BaseTestCase):
                     self.assertEqual(secret, f.read())
         os.unlink(infile.name)
         os.unlink(outfile.name)
+
+    def test_builds(self):
+        """Test builds subcommand"""
+        ZC = ZuulClient()
+        with self.assertRaisesRegex(Exception,
+                                    '--voting and --non-voting are '
+                                    'mutually exclusive'):
+            exit_code = ZC._main(
+                ['--zuul-url', 'https://fake.zuul',
+                 'builds', '--tenant', 'tenant1', '--voting', '--non-voting'])
+        with patch('requests.Session') as mock_sesh:
+            session = mock_sesh.return_value
+            session.get = MagicMock(
+                side_effect=mock_get(
+                    MagicMock(return_value=FakeRequestResponse(200, []))
+                )
+            )
+            exit_code = ZC._main(
+                ['--zuul-url', 'https://fake.zuul', 'builds',
+                 '--pipeline', 'gate',
+                 '--tenant', 'tenant1',
+                 '--change', '1234', '--job', 'job1', '--held'])
+            session.get.assert_any_call(
+                'https://fake.zuul/api/tenant/tenant1/builds',
+                params={'pipeline': 'gate',
+                        'change': '1234',
+                        'job_name': 'job1',
+                        'held': True,
+                        'skip': 0,
+                        'limit': 50}
+            )
+            self.assertEqual(0, exit_code)
+
+    def test_build_info(self):
+        """Test build-info subcommand"""
+        ZC = ZuulClient()
+        with self.assertRaisesRegex(Exception,
+                                    '--show-artifacts, --show-job-output and '
+                                    '--show-inventory are mutually exclusive'):
+            exit_code = ZC._main(
+                ['--zuul-url', 'https://fake.zuul',
+                 'build-info', '--tenant', 'tenant1',
+                 '--uuid', 'a1a1a1a1',
+                 '--show-artifacts', '--show-job-output'])
+        with patch('requests.Session') as mock_sesh:
+            session = mock_sesh.return_value
+            fakejson = {
+                'uuid': 'a1a1a1a1',
+                'job_name': 'tox-py38',
+                'result': 'SUCCESS',
+                'held': False,
+                'start_time': '2020-09-10T14:08:55',
+                'end_time': '2020-09-10T14:13:35',
+                'duration': 280.0,
+                'voting': True,
+                'log_url': 'https://log.storage/',
+                'node_name': None,
+                'error_detail': None,
+                'final': True,
+                'artifacts': [
+                    {'name': 'Download all logs',
+                     'url': 'https://log.storage/download-logs.sh',
+                     'metadata': {
+                         'command': 'xxx'}
+                    },
+                    {'name': 'Zuul Manifest',
+                     'url': 'https://log.storage/zuul-manifest.json',
+                     'metadata': {
+                         'type': 'zuul_manifest'
+                     }
+                    },
+                    {'name': 'Unit Test Report',
+                     'url': 'https://log.storage/testr_results.html',
+                     'metadata': {
+                         'type': 'unit_test_report'
+                     }
+                    }],
+                'provides': [],
+                'project': 'project1',
+                'branch': 'master',
+                'pipeline': 'check',
+                'change': 1234,
+                'patchset': '1',
+                'ref': 'refs/changes/34/1234/1',
+                'newrev': None,
+                'ref_url': 'https://gerrit/1234',
+                'event_id': '6b28762adfce415ba47e440c365ae624',
+                'buildset': {'uuid': 'b1b1b1'}}
+            session.get = MagicMock(
+                return_value=FakeRequestResponse(200, fakejson))
+            exit_code = ZC._main(
+                ['--zuul-url', 'https://fake.zuul',
+                 'build-info', '--tenant', 'tenant1',
+                 '--uuid', 'a1a1a1a1'])
+            session.get.assert_any_call(
+                'https://fake.zuul/api/tenant/tenant1/build/a1a1a1a1')
+            self.assertEqual(0, exit_code)
